// $Id$

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) Andreas Huber Doenni 2002-2004.
// Use, modification and distribution are subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////

//=============================================================================
/**
 *  @file    CarrierState.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2003-2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================


//////////////////////////////////////////////////////////////////////////////
// The following code implements the state-machine:
//
//
//                                                   O
//                                                   | *1
//                                                   V
//                                         +-----------------+
//                                         |     CARRIER     |
// +------------------------------------+--+-----------------+----------+-------------------------+
// |          CARRIER ID STATUS         :     CARRIER SLOT MAP STATUS   : CARRIER ACCESSING STATUS|
// |    O  +---------------------+      :                               :                         |
// |    | /                      |      :   O                           : O                       |
// |    V/  *2  +------------+   |      :*12|    +------------+         : |*17 +-----------+      |
// | +-(C)----->|   ID NOT   |   |      :   +--->|  SLOT MAP  |         : +--->|    NOT    |      |
// | |  |  +----|    READ    |   |      :    +---|  NOT READ  |         :      |  ACCESSED |      |
// | |  |  |    +------------+   |      :    |   +------------+         :      +-----------+      |
// | |*3|*4|*6     |*11   |*7    |*5    :    |          |               :            |            |
// | |  |  |       V      |      |      :    |*13       |*14            :            |*18         |
// | |  |  | +----(C)     |      |      :    |          V               :            |            |
// | |  |  | |     |*10   |      |      :    |   +------------+         :            V            |
// | |  |  | |     V      V      |      :    |   |   WAITING  |         :      +-----------+      |
// | |  |  | |  +------------+   |      :    |   |  FOR HOST  |         :      |    IN     |      |
// | +--|--|-|->|   WAITING  |   |      :    |   +------------+         :      |   ACCESS  |      |
// |    |  | |  |  FOR HOST  |   |      :    |     |        |           :      +-----------+      |
// |    |  | |  +------------+   |      :    |     |*15     |*16        :        |       |        |
// |    |  | |    |*8      |*9   |      :    V     V        V           :        |*19    |*20     |
// |    V  V V    V        V     V      : +------------+ +------------+ :        V       V        |
// |   +------------+    +------------+ : |  SLOT MAP  | |  SLOT MAP  | : +---------+ +---------+ |
// |   |     ID     |    |     ID     | : |VERIFICATION| |VERIFICATION| : | CARRIER | | CARRIER | |
// |   |VERIFICATION|    |VERIFICATION| : |     OK     | |    FAIL    | : | COMPLETE| | STOPPED | |
// |   |     OK     |    |   FAILED   | : +------------+ +------------+ : +---------+ +---------+ |
// |   +------------+    +------------+ :                               :                         |
// |                                    :                               :                         |
// +------------------------------------+-------------------------------+-------------------------+
//                                              |*21
//                                              v
//                                             (O)
//
//  *1: A carrier is instanciated.
//  *2: A "Bind" or "CarrierNotification" service is received.
//  *3: [Normal] A carrierID not currently existing at the equipment is successfully read.
//      Abnormal: A carrierID is read successfully but an equipment based verification field.
//  *4: [ID Read fail] A "ProcessWithCarrier" service is received.
//  *5: [ID Read fail] A "CancelCarrier" service is received.
//  *6: Carrier ID is read successfully and the equipment has verified the carrierID successfully.
//  *7: Carrier ID is read unsuccessfully.
//  *8: A "ProceedWithCarrier" service is received.
//  *9: A "CancelCarrier" service is received.
// *10: BypassReadID variable is set to FALSE, and carrier is received when the id reader is
//      not in service or not installed.
// *11: BypassReadID variable is set to TRUE, and carrier is received when the id reader is
//      not in service or not installed.
// *12: A carrier is instanciated.
// *13: Slot Map is read and verified successfully by the equipment.
// *14: [Normal host based verification]
//      Slot Map read successfully by the equipment and the equipment is waiting for host verification.
//      [Equipment based verification fail]
//      Slot Map read successfully but equipment based verification has failed.
//      [Slot Map read fail] Slot Map cannot be read.
//      [Abnormal substrate position with the carrier]
//      The Slot Map read has indicated an abnormal substrate position.
// *15: A "ProceedWithCarrier" service is received.
// *16: A "CancelCarrier" service is received.
// *17: A carrier object is instatiated.
// *18: The equipment starts accessing the carrier.
// *19: The equipment finishes accessing the carrier normally.
// *20: The equipment finishes accessing the carrier abnormally.
// *21: [Normal] The carrier is unloaded from the equipment.
//      [Abnormal by service]
//      "CancelBind" or "CancelCarrierNotification" service is received prior to the carrier load.
//      [Abnormal by the equipment]
//      An equipment based verification fails and the equipment performs a self-initiated
//      "CarrierBind" service.
//

#ifndef JYUGEM_CMS_CARRIERSTATE_H
#define JYUGEM_CMS_CARRIERSTATE_H

#ifdef _MSC_VER
#pragma warning( disable: 4267 )
#endif

#include "jyugem/bee/state_machine_task.h"
#include "jyugem/bee/viewed_state.h"
namespace bee = jyugem::bee;


struct EvBind                      : fsm::event< EvBind > {};
struct EvCancelAllCarrierOut       : fsm::event< EvCancelAllCarrierOut > {};
struct EvCancelBind                : fsm::event< EvCancelBind > {};
struct EvCancelCarrier             : fsm::event< EvCancelCarrier > {};
struct EvCancelCarrierToID         : fsm::event< EvCancelCarrierToID > {};
struct EvCancelCarrierToMap        : fsm::event< EvCancelCarrierToMap > {};
struct EvCancelCarrierAtPort       : fsm::event< EvCancelCarrierAtPort > {};
struct EvCancelCarrierNotification : fsm::event< EvCancelCarrierNotification > {};
struct EvCancelCarrierOut          : fsm::event< EvCancelCarrierOut > {};
struct EvCarrierIn                 : fsm::event< EvCarrierIn > {};
struct EvCarrierNotification       : fsm::event< EvCarrierNotification > {};
struct EvCarrierOut                : fsm::event< EvCarrierOut > {};
struct EvCarrierReCreate           : fsm::event< EvCarrierReCreate > {};
struct EvCarrierRelease            : fsm::event< EvCarrierRelease > {};
struct EvCarrierTagWriteData       : fsm::event< EvCarrierTagWriteData > {};
struct EvReceiveCarrier            : fsm::event< EvReceiveCarrier > {};
struct EvProceedWithCarrier        : fsm::event< EvProceedWithCarrier > {};
struct EvProceedWithCarrierToID    : fsm::event< EvProceedWithCarrierToID > {};
struct EvProceedWithCarrierToMap   : fsm::event< EvProceedWithCarrierToMap > {};
struct EvHostVerification          : fsm::event< EvHostVerification > {};
struct EvVerificationFail          : fsm::event< EvVerificationFail > {};
struct EvVerificationOK            : fsm::event< EvVerificationOK > {};
struct EvRead                      : fsm::event< EvRead > {};
struct EvSlotMapRead               : fsm::event< EvSlotMapRead > {};
struct EvAccess                    : fsm::event< EvAccess > {};
struct EvAbnormalPosition          : fsm::event< EvAbnormalPosition > {};
struct EvNormal                    : fsm::event< EvNormal > {};
struct EvAbnormal                  : fsm::event< EvAbnormal > {};


template< class T >
boost::intrusive_ptr< T > MakeIntrusive( T * p )
{
  return boost::intrusive_ptr< T >( p );
}

//
//
namespace jyugem { namespace cms
{

typedef bee::named_state     NamedState;
typedef bee::viewed_state<>  ViewedState;
typedef bee::viewed_state<1> SlotViewed;
typedef bee::viewed_state<2> AccessViewed;

//
// State Machine
//
class Carrier;
struct InitCarrierStatus;

class CarrierState : public state_machine_task< CarrierState, InitCarrierStatus >
{
public:
  typedef bee::state_machine_task< CarrierState, InitCarrierStatus >  task_base;

  // id verification status
  enum { STATUS_UNKNOWN = 0, STATUS_REQUEST, STATUS_OK, STATUS_FAIL };

  CarrierState(Carrier * carrier);
  ~CarrierState() {}

  template< class Event >
  void Repost( const Event & )
  {
    post_event( MakeIntrusive( new Event() ) );
  }

  virtual int send(trigger * trig)
  {
    return send_trigger<ViewedState>(trig->name());
  }

  const cstr_id * get_current_name(int region);

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  int  m_idread;
  int  m_verification;
  int  m_slotmap_verification;
};


//
// Innor State
//
struct IDStatus;
struct SlotMapStatus;
struct AccessingStatus;
struct InitIDStatus;
struct IDNotRead;
struct WaitingForHost;
struct IDVerificationOK;
struct IDVerificationFailed;
struct SlotMapNotRead;
struct SlotMapWaitingForHost;
struct SlotMapVerificationOK;
struct SlotMapVerificationFail;
struct NotAccessed;
struct InAccess;
struct CarrierComplete;
struct CarrierStopped;

//
//-----------------------------------------------------------------------------
//
struct InitCarrierStatus : public fsm::simple_state< InitCarrierStatus,
  CarrierState, fsm::no_reactions,
  mpl::list< IDStatus, SlotMapStatus, AccessingStatus > >,
  NamedState
{
  InitCarrierStatus() : NamedState() { TRACE_STATE_ENTER(); }
};


//
//-----------------------------------------------------------------------------
//
struct IDStatus : public fsm::simple_state< IDStatus,
  InitCarrierStatus::orthogonal<0>, fsm::no_reactions, InitIDStatus >,
  NamedState
{
  IDStatus() : NamedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct InitIDStatus : public fsm::state< InitIDStatus, IDStatus,
  fsm::custom_reaction< EvEvalCondition > >,
  NamedState
{
  InitIDStatus( my_context ctx) : NamedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    post_event( boost::intrusive_ptr< EvEvalCondition >( new EvEvalCondition() ) );
  }

  fsm::result react(const EvEvalCondition & )
  {
    if (! outermost_context().m_idread)
      return transit< IDNotRead >();
    else if (outermost_context().m_verification == CarrierState::STATUS_UNKNOWN)
      return transit< WaitingForHost >();
    else if (outermost_context().m_verification == CarrierState::STATUS_OK)
      return transit< IDVerificationOK >();
    else // if (outermost_context().m_verification == CarrierState::STATUS_FAIL)
      return transit< IDVerificationFailed >();
  }
};

//
//-----------------------------------------------------------------------------
//
struct IDNotRead : public fsm::state< IDNotRead, IDStatus,
  mpl::list< fsm::custom_reaction< EvRead >,
    fsm::custom_reaction< EvReceiveCarrier > > >,
  ViewedState
{
  IDNotRead( my_context ctx ) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();

    // Request to read Carrier ID

    bee::trigger * trig = new trigger("IDNotRead");
    outermost_context().changed(trig);
  }

  fsm::result react(const EvRead& )
  {
    if (outermost_context().m_verification)
      return transit< IDVerificationOK >();
    else
      return transit< WaitingForHost >();
  }

  fsm::result react(const EvReceiveCarrier& )
  {
    if (outermost_context().m_idread)
      return transit< IDVerificationOK >();
    else
      return transit< WaitingForHost >();
  }
};

//
//-----------------------------------------------------------------------------
//
struct WaitingForHost : public fsm::state< WaitingForHost, IDStatus,
  mpl::list< fsm::transition< EvProceedWithCarrier, IDVerificationOK,
    CarrierState, &CarrierState::Repost< EvProceedWithCarrier > >,
    fsm::transition< EvCancelCarrier, IDVerificationFailed,
    CarrierState, &CarrierState::Repost< EvCancelCarrier > > > >,
  ViewedState
{
  WaitingForHost( my_context ctx ) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("WaitingForHost");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct IDVerificationOK : public fsm::state< IDVerificationOK,
  IDStatus >, ViewedState
{
  IDVerificationOK( my_context ctx ) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("IDVerificationOK");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct IDVerificationFailed : public fsm::state< IDVerificationFailed,
  IDStatus >, ViewedState
{
  IDVerificationFailed( my_context ctx ) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("IDVerificationFailed");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct SlotMapStatus : public fsm::simple_state< SlotMapStatus,
  InitCarrierStatus::orthogonal<1>, fsm::no_reactions, SlotMapNotRead >,
  NamedState
{
  SlotMapStatus() : NamedState() { TRACE_STATE_ENTER(); }

};

//
//-----------------------------------------------------------------------------
//
struct SlotMapNotRead : public fsm::simple_state< SlotMapNotRead,
  SlotMapStatus, mpl::list< fsm::custom_reaction< EvSlotMapRead >,
    fsm::transition< EvHostVerification, SlotMapWaitingForHost >,
    fsm::transition< EvVerificationFail, SlotMapWaitingForHost >,
    fsm::transition< EvAbnormalPosition, SlotMapWaitingForHost > > >,
  SlotViewed
{
    SlotMapNotRead() : SlotViewed() { TRACE_STATE_ENTER(); }

    fsm::result react(const EvSlotMapRead& )
    {
        if (outermost_context().m_slotmap_verification)
            return transit< SlotMapVerificationOK >();
        else
            return transit< SlotMapWaitingForHost >();
    }
};

//
//-----------------------------------------------------------------------------
//
struct SlotMapWaitingForHost : public fsm::state<
  SlotMapWaitingForHost, SlotMapStatus,
  mpl::list< fsm::transition< EvProceedWithCarrier, SlotMapVerificationOK,
    CarrierState, &CarrierState::Repost< EvProceedWithCarrier > >,
    fsm::transition< EvCancelCarrier, SlotMapVerificationFail,
    CarrierState, &CarrierState::Repost< EvCancelCarrier > > > >,
  SlotViewed
{
  SlotMapWaitingForHost( my_context ctx ) : SlotViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("SlotMapWaitingForHost");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct SlotMapVerificationOK : public fsm::state<
  SlotMapVerificationOK, SlotMapStatus >,
  SlotViewed
{
  SlotMapVerificationOK( my_context ctx ) : SlotViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("SlotMapVerificationOK");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct SlotMapVerificationFail : public fsm::state<
  SlotMapVerificationFail, SlotMapStatus >,
  SlotViewed
{
  SlotMapVerificationFail( my_context ctx ) : SlotViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("SlotMapVerificationFail");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct AccessingStatus : public fsm::simple_state< AccessingStatus,
  InitCarrierStatus::orthogonal<2>, fsm::no_reactions, NotAccessed >,
  NamedState
{
  AccessingStatus() : NamedState() { TRACE_STATE_ENTER(); }

};

//
//-----------------------------------------------------------------------------
//
struct NotAccessed : public fsm::simple_state< NotAccessed, AccessingStatus,
  fsm::transition< EvAccess, InAccess > >,
  AccessViewed
{
  NotAccessed() : AccessViewed() { TRACE_STATE_ENTER(); }

};

//
//-----------------------------------------------------------------------------
//
struct InAccess : public fsm::state< InAccess, AccessingStatus,
  mpl::list< fsm::transition< EvNormal, CarrierComplete >,
    fsm::transition< EvAbnormal, CarrierStopped > > >,
  AccessViewed
{
  InAccess( my_context ctx ) : AccessViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("InAccess");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct CarrierComplete : public fsm::state< CarrierComplete,
  AccessingStatus >, AccessViewed
{
  CarrierComplete( my_context ctx ) : AccessViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("CarrierComplete");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct CarrierStopped : public fsm::state< CarrierStopped,
 AccessingStatus >, AccessViewed
{
  CarrierStopped( my_context ctx ) : AccessViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("CarrierStopped");
    outermost_context().changed(trig);
  }
};

} }  // *namespace* jyugem::cms

#endif  /* JYUGEM_CMS_CARRIERSTATE_H */
