// $Id$

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) Andreas Huber Doenni 2002-2004.
// Use, modification and distribution are subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////

//=============================================================================
/**
 *  @file    LoadPortState.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2003-2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================


//////////////////////////////////////////////////////////////////////////////
// The following code implements the state-machine:
//
//
// +-----------------------------------------------------------+
// |  O--->(H)                          LOAD PORT TRANSFER     |
// |    *1                                                     |
// |   +---------------------------------------------------+   |
// |   |                                                   |   |
// |   |                   OUT OF SERVICE                  |   |
// |   |                                                   |   |
// |   +---------------------------------------------------+   |
// |               *2|                       ^                 |
// |                 V                       | *3              |
// |   +---------------------------------------------------+   |
// |   |  O--->(C)          IN SERVICE                     |   |
// |   |    *4                                             |   |
// |   |  +---------------------------------------------+  |   |
// |   |  |  O--->(C)                                   |  |   |
// |   |  |    *5                                       |  |   |
// |   |  |  +------------+             +------------+  |  |   |
// |   |  |  |  READY TO  |             |  READY TO  |  |  |   |
// |   |  |  |    LOAD    |             |   UNLOAD   |  |  |   |
// |   |  |  +------------+             +------------+  |  |   |
// |   |  |     |    ^    TRANSFER READY    |     ^     |  |   |
// |   |  +-----|----|----------------------|-----|-----+  |   |
// |   |      *6|    |*8         ^        *7|     |*9      |   |
// |   |        V    |           |*10       V     |        |   |
// |   |  +---------------------------------------------+  |   |
// |   |  |                                             |  |   |
// |   |  |              TRANSFER BLOCKED               |  |   |
// |   |  |                                             |  |   |
// |   |  +---------------------------------------------+  |   |
// |   +---------------------------------------------------+   |
// +------------------+------------------+---------------------+
//
// *1: System reset
// *2: The host or operator has invoked the "ChangeServiceStatus" service for this load port
//     with a value of IN SERVICE.
// *3: The host or operator has invoked the "ChangeServiceStatus" service for this load port
//     with a value of OUT OF SERVICE.
// *4: [Service] The host or operator has invoked the "ChangeServiceStatus" service for this load port
//     with a value of IN SERVICE.
//     [System reset] This transition can be activated by an equipment re-initialization.
// *5: [Service] The host or operator has invoked the "ChangeServiceStatus" service for this load port
//     with a value of IN SERVICE.
//     [System reset] This transition can be activated by an equipment re-initialization.
//     [Failed Transfer] If a transfer fails, this transition is activated by transition #10
// *6: [Manual]
//     The equipment recognizes the logical indication of the start of a manual load transfer.
//     This trigger is configurable by the user.
//     [Automated] The PIO load transfer is beginning and the PIO "READY" signal activated.
//     [Internal buffer] A "CarrierOut" service has started for this load port.
// *7: [Manual]
//     The equipment recognizes the logical indication of the start of a manual load transfer.
//     This trigger is configurable by the user.
//     [Automated] The PIO load transfer is beginning and the PIO "READY" signal activated.
//     [Internal buffer] A "CarrierOut" service has started for this load port.
//     [By CarrierReCreate Service]
//     A "CarrierReCreate" service command has been issue by host or operator on a carrier
//     currently in the "CarrierComplete" or "CarrierStopped" state.
// *8: [Manual]
//     The carrier unload transfer has completed, and the load port is now empty and ready for
//     load transfer. This is indicated when two conditions are met, the presence signal indicates
//     that no carrier is present and the operator logically that the transfer is complete.
//     [Automated] The PIO unload transfer ends with the PIO "COMPT" signal.
//     [Internal buffer]
//     The carrier has finished its moved from the load port into the internal buffer,
//     and no "CarrierOut" services are queued for this load port.
// *9: [Manual]
//     Processing for substrates continued within the carrier has completed, or a
//     "CancelCarrier/CancelCarrierAtPort" service has been issued, and the carrier
//     has returned to the load/unload position on the load port.
//     [Automated] Processing for substrates belonging to the carrier has completed, or a
//     "CancelCarrier/CancelCarrierAtPort" service has been received, and the carrier
//     has returned to the load/unload position.
//     [Internal buffer]
//     A carrier has completed its moved from the internal buffer into the load port.
// *10: The transfer was unsuccessful, and the carrier was no loaded and unloaded.
//
//////////////////////////////////////////////////////////////////////////////
//
//                  +----------------------+
//                  |  LOAD PORT TRANSFER  |
// +----------------+-+------------------+-+-------------------+
// | O-->(H)          | O--+             | O---+               |
// |                  |    V             |     V               |
// |  +------------+  |  +------------+  |   +------------+    |
// |  |   MANUAL   |  |  |     NOT    |  |   |    NOT     |    |
// |  |            |  |  |  RESERVED  |  |   | ASSOCIATED |    |
// |  +------------+  |  +------------+  |   +------------+    |
// |      |    ^      |      |    ^      |       |   ^         |
// |    *1|    |*2    |    *3|    |*4    |     *5| *6|  +----+ |
// |      V    |      |      V    |      |       V   |  V  *7| |
// |  +------------+  |  +------------+  |   +------------+  | |
// |  |    AUTO    |  |  |  RESERVED  |  |   | ASSOCIATED |--+ |
// |  |            |  |  |            |  |   |            |    |
// |  +------------+  |  +------------+  |   +------------+    |
// |                  |     LOAD PORT    | LOAD PORT/CARRIER   |
// |   ACCESS MODE    |    RESERVATION   | ASSOCIATION         |
// +------------------+------------------+---------------------+
//
// *1: The host or operator has executed a "ChangeAccess" service with the value of AUTO.
//     This trigger can happen at anytime, except during a carrier transfer.
// *2: The host or operator has executed a "ChangeAccess" service with the value of MANUAL.
//     This trigger can happen at anytime, except during a carrier transfer.
// *3: [Service] If reserved by service, the host or operator sends a "ReserveAtPort" or a
//               "Bind" service to the production equipment.
//     [CarrierOut] This happens when the equipment physicaly initiates a "CarrierOut" operation.
// *4: [Service] If reservation is cancelled by service, the host or operator sends a
//               "CancelReserveAtPort" or a "CancelBind".
//     [CarrierOut] A carrier arrived at the reserved port.
// *5: [Service Normal]
//     If associated by service in the normal situation, host sends a "Bind" service to
//     the production equipment when the port is unoccupied.
//     [Service Abnormal]
//     If associated with a service in an abnormal situation, host sends a "ProceedWithCarrier"
//     service to the production equipment when the port is unoccupied.
//     [CarrierID Read]
//     If associated by a "CarrierID" read, the production equipment creates the association
//     at the time the "CarrierID" read is performed.
//     [Known Carrier] A carrier already known to the production equipment is being loaded
//     onto the load port.
//     This happens when the CarrierOut service is initiated.
// *6: [Service] If cancellation of a load port association is required; then, this
//     can be accomplished by sending a "CancelBind" service to the production equipment
//     before the carrier arrives to the load port or before a transfer sequence has started.
//     [Carrier Unload]
//     An association cancellation may also performed by removing the carrier from the load
//     port or by the production equipment moving a carrier to an internal buffer position.
// *7: Production equipment based carrier verification fails, and the carrier assumes
//     the ID value from the carrier that on the load port.
//     [Internal buffer]
//     A carrier is unloaded and a queued "CarrierOut" service starts.
//
//

#ifndef JYUGEM_CMS_LOADPORTSTATE_H
#define JYUGEM_CMS_LOADPORTSTATE_H

#ifdef _MSC_VER
#pragma warning( disable: 4267 )
#endif

#include "jyugem/bee/state_machine_task.h"
#include "jyugem/bee/viewed_state.h"

namespace bee = jyugem::bee;


struct EvChangeServiceStatus       : fsm::event< EvChangeServiceStatus >
{
  EvChangeServiceStatus(int par = OUT_OF_SERVICE) : m_param(par) {}
  
  enum { OUT_OF_SERVICE, IN_SERVICE };
  int m_param;
};
struct EvReady                     : fsm::event< EvReady > {};
struct EvUnloadComplete            : fsm::event< EvUnloadComplete > {};
struct EvProcComplete              : fsm::event< EvProcComplete > {};
struct EvTransferFail              : fsm::event< EvTransferFail > {};
struct EvChangeAccess              : fsm::event< EvChangeAccess >
{
  EvChangeAccess(int m = MANUAL) : m_mode(m) {}

  enum { MANUAL, AUTO };
  int m_mode;
};
struct EvBind                      : fsm::event< EvBind > {};
struct EvCancelAllCarrierOut       : fsm::event< EvCancelAllCarrierOut > {};
struct EvCancelBind                : fsm::event< EvCancelBind > {};
struct EvCancelCarrier             : fsm::event< EvCancelCarrier > {};
struct EvCancelCarrierAtPort       : fsm::event< EvCancelCarrierAtPort > {};
struct EvCancelCarrierNotification : fsm::event< EvCancelCarrierNotification > {};
struct EvCancelCarrierOut          : fsm::event< EvCancelCarrierOut > {};
struct EvCancelReservationAtPort   : fsm::event< EvCancelReservationAtPort > {};
struct EvCarrierIn                 : fsm::event< EvCarrierIn > {};
struct EvCarrierNotification       : fsm::event< EvCarrierNotification > {};
struct EvCarrierOut                : fsm::event< EvCarrierOut > {};
struct EvCarrierReCreate           : fsm::event< EvCarrierReCreate > {};
struct EvCarrierRelease            : fsm::event< EvCarrierRelease > {};
struct EvCarrierTagWriteData       : fsm::event< EvCarrierTagWriteData > {};
struct EvProceedWithCarrier        : fsm::event< EvProceedWithCarrier > {};
struct EvReserveAtPort             : fsm::event< EvReserveAtPort > {};
struct EvVerificationFail          : fsm::event< EvVerificationFail > {};

//
//
namespace jyugem { namespace cms
{

typedef bee::named_state     NamedState;
typedef bee::viewed_state<>  ViewedState;
typedef bee::viewed_state<1> AccessViewed;
typedef bee::viewed_state<2> ReservationViewed;
typedef bee::viewed_state<3> AssociationViewed;

//
// State Machine
//
class  LoadPort;
struct InitLoadPortStatus;

class LoadPortState : public state_machine_task< LoadPortState, InitLoadPortStatus >
{
public:
  typedef bee::state_machine_task< LoadPortState, InitLoadPortStatus >  task_base;

  LoadPortState(LoadPort * ldp);
  ~LoadPortState() {}

  virtual int send(trigger * trig)
  {
    return send_trigger<ViewedState>(trig->name());
  }

  const cstr_id * get_current_name(int region);

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  bool m_blocked;
  bool m_loaded;
};

//
// Innor State
//
struct LoadPortTransfer;
struct AccessMode;
struct Reservation;
struct CarrierAssociation;
struct OutOfService;
struct InService;
struct TransferReady;
struct TransferBlocked;
struct ReadyToLoad;
struct Manual;
struct Auto;
struct NotReserved;
struct Reserved;
struct NotAssociated;
struct Associated;

//
//-----------------------------------------------------------------------------
//
struct InitLoadPortStatus : public fsm::simple_state< InitLoadPortStatus,
  LoadPortState, fsm::no_reactions, mpl::list< LoadPortTransfer, AccessMode,
  Reservation, CarrierAssociation > >,
  NamedState
{
  InitLoadPortStatus() : NamedState() { TRACE_STATE_ENTER(); }
};

//-----------------------------------------------------------------------------
struct LoadPortTransfer : public fsm::simple_state< LoadPortTransfer,
  InitLoadPortStatus::orthogonal<0>, fsm::no_reactions,
  mpl::list< fsm::deep_history< OutOfService > >, fsm::has_deep_history >,
  NamedState
{
  LoadPortTransfer() : NamedState()
  {
    TRACE_STATE_ENTER();
  }
};

//-----------------------------------------------------------------------------
struct InService : public fsm::state< InService, LoadPortTransfer,
  fsm::custom_reaction< EvChangeServiceStatus >,
  mpl::list< fsm::shallow_history< TransferReady > >, fsm::has_shallow_history >,
  NamedState
{
  InService(my_context ctx) : NamedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("InService");
    outermost_context().changed(trig);
  }    

  fsm::result react(EvChangeServiceStatus ce)
  {
    if (ce.m_param == EvChangeServiceStatus::OUT_OF_SERVICE)
      return transit< OutOfService >();
    else
      return discard_event() ;
  }

};

//-----------------------------------------------------------------------------
struct TransferCondition;

struct TransferReady : public fsm::state< TransferReady, InService,
  fsm::no_reactions, mpl::list< fsm::shallow_history< TransferCondition > >,
  fsm::has_shallow_history >,
  ViewedState
{
  TransferReady(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
  }

};

struct TransferCondition : public fsm::state< TransferCondition, TransferReady,
  fsm::custom_reaction< EvEvalCondition > >,
  NamedState
{
  TransferCondition(my_context ctx) : NamedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    post_event(boost::intrusive_ptr< EvEvalCondition >(new EvEvalCondition()));
  }

  fsm::result react( const EvEvalCondition & );
};


//-----------------------------------------------------------------------------
struct ReadyToLoad : public fsm::state< ReadyToLoad, TransferReady,
  fsm::transition< EvReady, TransferBlocked > >,
  ViewedState
{
  ReadyToLoad(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("ReadyToLoad");
    outermost_context().changed(trig);
  }

};

//-----------------------------------------------------------------------------
struct ReadyToUnload : public fsm::state< ReadyToUnload, TransferReady,
  fsm::transition< EvReady, TransferBlocked > >,
  ViewedState
{
  ReadyToUnload(my_context ctx) : ViewedState(), my_base(ctx) 
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("ReadyToUnload");
    outermost_context().changed(trig);
  }
};

//-----------------------------------------------------------------------------
struct OutOfService : public fsm::state< OutOfService, LoadPortTransfer, 
  fsm::custom_reaction< EvChangeServiceStatus > >,
  ViewedState
{
  OutOfService(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("OutOfService");
    outermost_context().changed(trig);
  }
  
  fsm::result react(EvChangeServiceStatus ce)
  {
    if (ce.m_param == EvChangeServiceStatus::IN_SERVICE)
      return transit< InService >();
    else
      return discard_event() ;
  }

};

//
//-----------------------------------------------------------------------------
//
struct TransferBlocked : public fsm::state< TransferBlocked, InService,
  mpl::list< fsm::custom_reaction< EvUnloadComplete >,
    fsm::custom_reaction< EvProcComplete >,
    fsm::custom_reaction< EvCancelCarrier >,
    fsm::custom_reaction< EvTransferFail > > >,
  ViewedState
{
  TransferBlocked(my_context ctx) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    outermost_context().m_blocked = true;
    bee::trigger * trig = new trigger("TransferBlocked");
    outermost_context().changed(trig);
  }

  ~TransferBlocked()
  {
    outermost_context().m_blocked = false;
  }

  fsm::result react(const EvUnloadComplete & )
  {
    outermost_context().m_loaded = false;
    return transit< ReadyToLoad >();
  }

  fsm::result react(const EvCancelCarrier & )
  {
    return transit< ReadyToUnload >();
  }

  fsm::result react(const EvProcComplete & )
  {
    return transit< ReadyToUnload >();
  }

  fsm::result react(const EvTransferFail & )
  {
    return transit< TransferReady >();
  }
};

//------------------------------------------------------------------------------
// Load Port Access Mode Status
//
struct AccessMode : public fsm::simple_state< AccessMode,
  InitLoadPortStatus::orthogonal<1>, fsm::no_reactions,
  mpl::list< fsm::shallow_history< Manual > >, fsm::has_shallow_history >,
  NamedState
{
  AccessMode() : NamedState()
  {
    TRACE_STATE_ENTER();
  }
};

//
struct Manual : fsm::state< Manual, AccessMode,
  fsm::custom_reaction< EvChangeAccess > >,
  AccessViewed
{
  Manual(my_context ctx) : AccessViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER(); 
    bee::trigger * trig = new trigger("Manual");
    outermost_context().changed(trig);
  }

  fsm::result react(EvChangeAccess ae)
  {
    if (ae.m_mode == EvChangeAccess::AUTO)
      return transit< Auto >();
    else
      return discard_event() ;
  }
};

//
struct Auto : public fsm::state< Auto, AccessMode,
  fsm::custom_reaction< EvChangeAccess > >,
  AccessViewed
{
  Auto(my_context ctx) : AccessViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("Auto");
    outermost_context().changed(trig);
  }

  fsm::result react(EvChangeAccess ae)
  {
    if (ae.m_mode == EvChangeAccess::MANUAL)
      return transit< Manual >();
    else
      return discard_event() ;
  }
};

//------------------------------------------------------------------------------
// Load Port Reservation Status
//
struct Reservation : public fsm::simple_state< Reservation,
  InitLoadPortStatus::orthogonal<2>, fsm::no_reactions, NotReserved >,
  NamedState
{
  Reservation() : NamedState()
  {
    TRACE_STATE_ENTER();
  }
};

struct NotReserved : public fsm::state< NotReserved, Reservation,
  mpl::list< fsm::transition< EvReserveAtPort, Reserved >,
    fsm::transition< EvBind, Reserved > > >,
  ReservationViewed
{
  NotReserved(my_context ctx) : ReservationViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER(); 
    bee::trigger * trig = new trigger("NotReserved");
    outermost_context().changed(trig);
  }
};

struct Reserved : public fsm::state< Reserved, Reservation,
  mpl::list< fsm::transition< EvCancelReservationAtPort, NotReserved >,
    fsm::transition< EvCancelBind, NotReserved > > >,
  ReservationViewed
{
  Reserved(my_context ctx) : ReservationViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER(); 
    bee::trigger * trig = new trigger("Reserved");
    outermost_context().changed(trig);
  }
};

//------------------------------------------------------------------------------
// Load Port/ Carrier Association Status
//
struct CarrierAssociation : public fsm::simple_state< CarrierAssociation,
  InitLoadPortStatus::orthogonal<3>, fsm::no_reactions, NotAssociated >,
  NamedState
{
  CarrierAssociation() : NamedState()
  {
    TRACE_STATE_ENTER(); 
  }
};

//
struct NotAssociated : public fsm::state< NotAssociated,
  CarrierAssociation, mpl::list< fsm::transition< EvBind, Associated >,
    fsm::transition< EvCarrierOut, Associated >,
    fsm::transition< EvProceedWithCarrier, Associated > > >,
  AssociationViewed
{
  NotAssociated(my_context ctx) : AssociationViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER(); 
    bee::trigger * trig = new trigger("NotAssociated");
    outermost_context().changed(trig);
  }
};

//
struct Associated : public fsm::state< Associated,
  CarrierAssociation, mpl::list<
    fsm::transition< EvCancelBind, NotAssociated >,
    fsm::transition< EvCarrierRelease, NotAssociated >,
    fsm::deferral< EvVerificationFail > > >,
  AssociationViewed
{
  Associated(my_context ctx) : AssociationViewed(), my_base( ctx )
  {
    TRACE_STATE_ENTER(); 
    bee::trigger * trig = new trigger("Associated");
    outermost_context().changed(trig);
  }
};


} }  // *namespace* jyugem::cms

#endif  // JYUGEM_CMS_LOADPORTSTATE_H
