// $Id$

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) Andreas Huber Doenni 2002-2004.
// Use, modification and distribution are subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////

//=============================================================================
/**
 *  @file    ProcessJob.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2003-2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================


//////////////////////////////////////////////////////////////////////////////
// The following code implements the state-machine:
//
// E40 : STANDARD FOT PROCESS MANAGEMENT
//
// Process Job State Model
//
// +------------------------------------------------------------------------+
// |                                                        PROCESSJOB      |
// |      O     +---------------------------------------------------------+ |
// |   *1 |     |                                                 ACTIVE  | |
// |      |     |  +----------------------------+      +------------+     | |
// |      V     |  |                EXECUTING   | *11  |  STOPPING  |----------+
// | +--------+ |  | *2  +----------------+     |----->|            |---+ | |  |
// | | QUEUED/|--------->|   SETTING UP   |     |      +------------+   | | |  |
// | | POOLED | |  |  +--|                |     |            ^          | | |  |
// | +--------+ |  |  |  +----------------+     |            |*12       | | |  |
// |      |     |  |  |*4        |*3            |  +------------------+ | | |  |
// |      |     |  |  |          V              |  |            PAUSE | | | |  |
// |      |     |  |  |  +----------------+     |*8|   +-----------+  | | | |  |
// |      |     |  |  |  |  WAITING FOR   |     |----->|  PAUSING  |  | | | |  |
// |      |     |  |  |  |     START      |     |  |   |           |  | | | |  |
// |      |     |  |  |  +----------------+     |  |   +-----------+  | | | |  |
// |      |     |  |  |          |*5       (H)<----|         |*9      | | | |  |
// |      |     |  |  |          V           *10|  |         V        | | | |  |
// |      |     |  |  |  +----------------+     |  |   +-----------+  | | | |  |
// |      |     |  |  +->|   PROCESSING   |     |  |   |  PAUSED   |  | | | |  |
// |      |     |  |     |                |     |  |   |           |  | | | |  |
// |      |     |  |     +----------------+     |  |   +-----------+  | | | |  |
// |      |     |  |             |*6            |  +------------------+ | | |  |
// |      |     |  |             V              |            |*15    *14| | |  |
// |      |     |  |     +----------------+     |            V          | | |  |
// |      |     |  |     |    PROCESS     |     | *13  +------------+   | | |  |
// |      |     |  |     |    COMPLETE    |     |----->|  ABORTING  |<--+ | |  |
// |      |     |  |     +----------------+     |      |            |     | |  |
// |      |     |  |             |*7            |      +------------+     | |  |
// |      |     |  +-------------|--------------+            |            | |  |
// |      |     +----------------|---------------------------|------------+ |  |
// +------|----------------------|---------------------------|--------------+  |
//        |                      |                           |*16              |
//        |     *18              V                           |                 |
//        +--------------------> O <-------------------------+                 |
//                               ^                             *17             |
//                               +---------------------------------------------+
//
//
//  *1: The processing resource accepts a ProcessJob create request.
//  *2: The processing resource has been allocated to the ProcessJob.
//  *3: Job material is present and the processing resource is ready to start
//      the ProcessJob and PRProcessStart attribute is not set.
//  *4: Material is present and ready for processing. PRProcessStart attribute is set.
//  *5: Job start directive.
//  *6: Material processing completed.
//  *7: Job material departed the processing resource and resource post-conditioning
//      completed, or superceded by another process job on the same material.
//  *8: The processing resource initiated a process pause action. (it received a
//      "Pause" command or initiated an internal pause)
//  *9: The processing resource paused the job.
// *10: The processing resource resumed the job.
// *11: The processing resource initiated a process stop action. (it received a
//      "Stop" command or initiated an internal stop)
// *12: The processing resource initiated a process stop action. (it received a
//      "Stop" command or initiated an internal stop)
// *13: The processing resource initiated a process abort action. (it received a
//      "Abort" command or initiated an internal abort)
// *14: The processing resource initiated a process abort action. (it received a
//      "Abort" command or initiated an internal abort)
// *15: The processing resource initiated a process abort action. (it received a
//      "Abort" command or initiated an internal abort)
// *16: The processing resource abort procedure is complete and for some processing
//      equipment the related substrates are moved out as part of the error recovery.
// *17: The processing resource stop procedure is complete.
// *18: "Cancel", "Stop" or "Abort" command received.
//
//

#ifndef JYUGEM_PJM_PROCESSJOBSTATE_H
#define JYUGEM_PJM_PROCESSJOBSTATE_H

#ifdef _MSC_VER
#pragma warning( disable: 4267 )
#endif

#include "jyugem/bee/state_machine_task.h"
#include "jyugem/bee/viewed_state.h"
namespace bee = jyugem::bee;

struct EvAllocate  : fsm::event< EvAllocate > {};   //
struct EvStart     : fsm::event< EvStart > {};      //
struct EvCompleted : fsm::event< EvCompleted > {};  //
struct EvDepart    : fsm::event< EvDepart > {};     //
struct EvPause     : fsm::event< EvPause > {};      //
struct EvPaused    : fsm::event< EvPaused > {};     //
struct EvResume    : fsm::event< EvResume > {};     //
struct EvStop      : fsm::event< EvStop > {};       //
struct EvStopped   : fsm::event< EvStopped > {};    //
struct EvAbort     : fsm::event< EvAbort > {};      //
struct EvAborted   : fsm::event< EvAborted > {};    //
struct EvCancel    : fsm::event< EvCancel > {};     //


//
// State classes
//
namespace jyugem { namespace pjm
{

typedef bee::named_state    NamedState;
typedef bee::viewed_state<> ViewedState;

//
// State Machine
//
class ProcessJob;
struct QueuedAndPooled;
class ProcessJobState : public bee::state_machine_task< ProcessJobState, QueuedAndPooled >
{
public:
  typedef bee::state_machine_task< ProcessJobState, QueuedAndPooled >  task_base;

  ProcessJobState(ProcessJob * pjm);
  ~ProcessJobState() {}

  virtual int send(trigger * trig)
  {
    return send_trigger<ViewedState>(trig->name());
  }

  // const std::string& current() { return this->current_state_name<ViewedState>(); }

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  bool m_PRProcessStart;
};


//
// Innor State
//
struct SettingUp;
struct Executing;
struct Pausing;
struct Stopping;
struct Aborting;
struct WaitingForStart;
struct Processing;
struct ProcessCompleted;
struct Paused;

//
//-----------------------------------------------------------------------------
//
struct Departed : fsm::state< Departed, ProcessJobState >, ViewedState
{
  Departed(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("Departed");
    outermost_context().changed(trig);
  }
};


//
//-----------------------------------------------------------------------------
//
struct QueuedAndPooled : fsm::simple_state< QueuedAndPooled, ProcessJobState,
  mpl::list< fsm::transition< EvAllocate, SettingUp >,
  fsm::transition< EvAbort, Departed >, fsm::transition< EvCancel, Departed >,
  fsm::transition< EvStop, Departed > > >,
  ViewedState
{
  QueuedAndPooled() : ViewedState()
  {
    TRACE_STATE_ENTER();
  }
};


//
//-----------------------------------------------------------------------------
//
struct Active : fsm::simple_state< Active, ProcessJobState,
  fsm::no_reactions, Executing >, NamedState
{
  Active() : NamedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct Executing : fsm::simple_state< Executing, Active, mpl::list<
  fsm::transition< EvPause, Pausing >, fsm::transition< EvStop, Stopping >,
  fsm::transition< EvAbort, Aborting > >,
  SettingUp, fsm::has_shallow_history >, NamedState
{
  Executing() : NamedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct SettingUp : fsm::state< SettingUp, Executing,
  fsm::custom_reaction< EvEvalCondition > >, ViewedState
{
  SettingUp( my_context ctx ) : ViewedState(), my_base( ctx )
  {
    TRACE_STATE_ENTER();
    post_event(boost::intrusive_ptr< EvEvalCondition >(new EvEvalCondition()));
  }

  fsm::result react( const EvEvalCondition & )
  {
    if (outermost_context().m_PRProcessStart)
      return transit< Processing >();
    else
      return transit< WaitingForStart >();
  }
};

//
//-----------------------------------------------------------------------------
//
struct WaitingForStart : fsm::state< WaitingForStart,
  Executing, fsm::transition< EvStart, Processing > >, ViewedState
{
  WaitingForStart(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new bee::trigger("Job Waiting For Start");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct Processing : fsm::state< Processing, Executing,
  fsm::transition< EvCompleted, ProcessCompleted > >, ViewedState
{
  Processing(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new bee::trigger("Processing");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct ProcessCompleted : fsm::state< ProcessCompleted,
  Executing, fsm::transition< EvDepart, Departed > >, ViewedState
{
  ProcessCompleted(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new bee::trigger("Completed");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
struct Stopping : fsm::simple_state< Stopping, Active,
  fsm::transition< EvStopped, Departed > >, ViewedState
{
  Stopping() : ViewedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct Aborting : fsm::simple_state< Aborting, Active,
  fsm::transition< EvAborted, Departed > >, ViewedState
{
  Aborting() : ViewedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct Pause : fsm::simple_state< Pause, Active,
  mpl::list< fsm::transition< EvResume, fsm::shallow_history< SettingUp > >,
    fsm::transition< EvStop, Stopping >,
    fsm::transition< EvAbort, Aborting > >, Pausing >, NamedState
{
  Pause() : NamedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
struct Pausing : fsm::simple_state< Pausing, Pause,
  fsm::transition< EvPaused, Paused > >, ViewedState
{
  Pausing() : ViewedState() { TRACE_STATE_ENTER(); }
};


//
//-----------------------------------------------------------------------------
//
struct Paused : fsm::simple_state< Paused, Pause >, ViewedState
{
    Paused() : ViewedState() { TRACE_STATE_ENTER(); }
};


} }  // *namespace* jyugem::pjm

#endif  /* JYUGEM_PJM_PROCESSJOBSTATE_H */
